/*
   Copyright 2009 Andrea Pagnani and Alfredo Braunstein

   This file is part of DILPERC (DILuted discrete PERCeptron analysis tool).

   DILPERC is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License Version 2 
   as published by the Free Software Foundation 

   DILPERC is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with DILPERC; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <math.h>
#include <unistd.h>
#include <stdlib.h>
#include <omp.h>
#include "random.h"



#define Hmax 20
#define SMALL 1e-20 
#define INF 1e+20

#define INVSQRT2 .70710678118654752440

/* definite integral in (-inf,x) of Normal(mu,sigma) */
#define normal(mu,invsqrtsigma,x) (.5*(1+erf(((x)-(mu))*invsqrtsigma*INVSQRT2)))


/* factor graph and messages*/
double *** mes;
double *** lmes;
double ** P;
int **zH;
double ** H;
omp_lock_t * locker;


double ** muv, **sigmav, ** auxv;

int Q,N,M;
int * perm;
double **X;
double *Y;

/* diluition parameter */
int nmax = 0;
int deltan = 0;
double *Hext;
double hdil = 0;
double beta = 1;
double deltahdil = 0;


/* convergence tolerance threshold */
double err = 1e-5;

/* structure for sorting fields used for decimation */
struct val_index {
	double val;
	int index;
};

int converge(int);
int nonlinear = 1;

void free_memory(void)
{
	int n,m;
	int NUM = omp_get_max_threads();

	free(Y);
	free(Hext);
	free(perm);
	free(locker);
	for (m = 0; m < M; ++m){
		for(n=0; n<N; ++n){
			free(mes[m][n]);
			free(lmes[m][n]);
		}
		free(X[m]);
		free(mes[m]);
		free(lmes[m]);
	}
	free(X);
	free(mes);
	free(lmes);


	for(n=0; n<N; ++n){
		free(P[n]);
		free(H[n]);
		free(zH[n]);
	}

	free(H);
	free(P);
	free(zH);

	for( n = 0; n < NUM; ++n ){
		free(muv[n]);
		free(sigmav[n]);
		free(auxv[n]);
	}

	free(muv);
	free(sigmav);
	free(auxv);
}



void init()
{
	int n, m, q, q2,i;
	int NUM = omp_get_max_threads();
	double norm;

	Y = calloc(M, sizeof(double));
	X = calloc(M, sizeof(double *));

	for (m = 0; m < M; ++m)
		X[m] = calloc(N, sizeof(double));

	mes = calloc(M, sizeof(double **));
	lmes = calloc(M, sizeof(double **));


	Hext = calloc(N, sizeof(double));

	perm = calloc(M, sizeof(int));

	P = calloc(N, sizeof(double *));

	H = calloc(N, sizeof(double *));
	locker = calloc(N, sizeof(omp_lock_t));
	zH = calloc(N,sizeof(int *));
	

	muv = calloc(NUM, sizeof(double *));
	sigmav = calloc(NUM, sizeof(double *));
	auxv = calloc(NUM, sizeof(double *));

	for( i = 0; i < NUM; ++i ){
		muv[i] = calloc(N, sizeof(double));
		sigmav[i] = calloc(N, sizeof(double));
		auxv[i] = calloc(2*Q+1, sizeof(double));
	}

	for (n = 0; n < N; ++n) {
		omp_init_lock(locker + n);
		H[n] = calloc(2*Q+1, sizeof(double));
		P[n] = calloc(2*Q+1, sizeof(double));
		zH[n] = calloc(2*Q+1,sizeof(int));
		Hext[n] = hdil;
		norm = 0;
		for (q = 0, q2 = -Q; q2 <= Q ; ++q, ++q2) {
			H[n][q] = 0;
			zH[n][q] = 0;
		}
	}

	for (m = 0; m < M; ++m) {
		perm[m] = m;
		mes[m] = calloc(N, sizeof(double *));
		lmes[m] = calloc(N, sizeof(double *));
		for (n = 0; n < N; ++n) {
			mes[m][n] = calloc(2*Q+1, sizeof(double));
			lmes[m][n] = calloc(2*Q+1, sizeof(double));
			for (q = 0; q < 2*Q+1; ++q){
				mes[m][n][q] = 1./(2*Q+1);
				lmes[m][n][q] = log(mes[m][n][q]);
				H[n][q] += lmes[m][n][q];
			}
		}
	}			


}


void read_matrix(char *filename)
{
	int n,m,aux;
	FILE *fp;


	
	fp = fopen( filename , "r");
	if (fp == NULL) {
		fprintf(stderr, "datafile file  %s doesn't exist ... exiting\n", filename);
		exit(-1);
	}

	for (m = 0; m < M; ++m) {
		aux = fscanf(fp,"%lf ", &Y[m]);
		for (n = 0; n < N; ++n) {
			aux = fscanf(fp,"%lf ", &X[m][n]);
		}
	}
	fclose(fp);
}


double max(double a, double b) 
{
	return a < b ? b : a;
}


void check_message(int n)
{

	int m,q;
	double norm =0;
	double * th =  auxv[omp_get_thread_num()];

	for (q = 0; q < 2*Q+1; ++q )
		th[q] = 1;
	
	for (m = 0; m < M; ++m)
		for (q = 0; q < 2 * Q + 1; ++q )
			th[q] *= mes[m][n][q];
	
	for (q = 0; q < 2 * Q + 1; ++q )
		norm += th[q];
	
	for (q = 0; q < 2 * Q + 1; ++q )
		printf("%i true %g running %g diff %g\n", q,th[q]/norm, H[n][q], th[q]/norm - H[n][q]);

}


double * get_message(int m, int n) 
{
	int q,q2;
	double * u = auxv[omp_get_thread_num()];
	double sup = -INF;
	double norm = 0;

	omp_set_lock(locker + n);
	
	for (q = -Q, q2 = 0; q <= Q; ++q, ++q2) {
		if (zH[n][q2] == 0)
			u[q2] = H[n][q2] - lmes[m][n][q2];
		else if (zH[n][q2] == 1)
			u[q2] = mes[m][n][q2] > SMALL ? -INF : H[n][q2];
		else 
			u[q2] = -INF;

		u[q2] += (q == 0) * Hext[n];  
		sup = max(sup, u[q2]);		
	}
	

	
	norm = 0;
	for (q = 0; q < 2 * Q + 1; ++q ){
		u[q] = exp( u[q] - sup );
		norm += u[q];
	}

	norm = 1./norm;
	for (q = 0; q < 2 * Q + 1; ++q )
		u[q] *= norm;

	omp_unset_lock(locker + n);

	return u;
}


void put_message(int m, int n, double  * u) 
{
	int q, q2;
	omp_set_lock(locker + n);
	for (q = -Q, q2 = 0; q <= Q; ++q, ++q2) {
		if (mes[m][n][q2] < SMALL)
			zH[n][q2]--;
		else
			H[n][q2] -= lmes[m][n][q2];


		if (u[q2] < SMALL)
			zH[n][q2]++;
		else {
			H[n][q2] += (lmes[m][n][q2] = log(u[q2]));
		}

		mes[m][n][q2] = u[q2];
	}			
	omp_unset_lock(locker + n);
}


/* updates pattern m and return message difference */
double update(int m)
{
	int n,q,q2;
	double mom1 = 0, mom2 = 0;
	double Smu = 0, Ssigma = 0, Cmu, Csigma;
	double A,diff,norm, udbss, delta;
	double * u = auxv[omp_get_thread_num()];
	double * sigma = sigmav[omp_get_thread_num()];
	double * mu = muv[omp_get_thread_num()];
	
/*	printf("(%d) ", omp_get_thread_num());*/

	/* # pragma omp parallel for default(none) private(n,q,q2,mom1,mom2,Smu,Ssigma) shared(N,u,mu,sigma,X) */
	for (n = 0; n < N; ++n) {
		mom1 = 0;
		mom2 = 0;
		u = get_message(m, n);
		for (q = 0, q2 = -Q; q2 <= Q; ++q, ++q2) {
			mom1 += u[q] * q2;
			mom2 += u[q] * q2 * q2;
		}
		mu[n] = X[m][n] * mom1;
		sigma[n] = X[m][n] * X[m][n] * (mom2 - mom1 * mom1);
		Smu += mu[n];
		Ssigma += sigma[n];
	}

	diff = 0;
	for (n = 0; n < N; ++n) {
		norm = 0;
		for (q = -Q, q2 = 0; q <= Q; ++q, ++q2) {
			Cmu = Smu - mu[n];
			Csigma = Ssigma - sigma[n];
			if (nonlinear) {
				A = normal(Cmu, 1 / sqrt(Csigma), -q * X[m][n]);
				u[q2] = A * Y[m] + (1 - A) * (1 - Y[m]);
			} else {
				udbss = 1 + 2 * beta * Csigma;
				delta = -Y[m] + q * X[m][n];
				u[q2] = exp(-(beta * (delta + Cmu) * (delta + Cmu)) / udbss) / sqrt(udbss);
			}
			norm += u[q2];

		}
		if (norm > SMALL) {
			for (q = -Q, q2 = 0; q <= Q; ++q, ++q2) {
				u[q2] /= norm;
				diff = max(diff, fabs(u[q2] - mes[m][n][q2]));
			}
		} else {
			for (q = -Q, q2 = 0; q <= Q; ++q, ++q2){ 
				u[q2] = 1./(2*Q+1);
				diff = max(diff, fabs(u[q2] - mes[m][n][q2]));
			}
		}
		put_message(m,n,u);
	}
	
	return diff;
}



double iterate() 
{
	int m,p,m2;
	double diff = 0, d;
	for (m = 0; m < M; ++m) {
		m2 = randint(M - m) + m;
                p = perm[m2];
                perm[m2] = perm[m];
                perm[m] = p;	      
	}
#pragma omp parallel for private(m,d) shared(perm) schedule(dynamic)
	for (m = 0; m < M; ++m) {
		d = update(perm[m]);
#pragma omp critical
		diff = max(diff, d);
	}
	return diff;
}


void print_marginals() 
{
	int n,q,q2;

	for( n = 0; n<N; ++n ){
		printf("Marg %i ",n);	
		for (q = 0, q2 = -Q; q2 <= Q; ++q, ++q2) 
			printf("%g ",P[n][q]);
		printf("\n");
	}

}


void compute_marginals()
{
	int n,q,q2;
	double norm, sup;

	for (n = 0; n < N; ++n) {
		sup = -INF;
		norm = 0;
		for (q = 0, q2 = -Q; q2 <= Q; ++q, ++q2) 
			sup = max(sup, H[n][q] + (q2 == 0) * Hext[n]);

		for (q = 0, q2 = -Q; q2 <= Q; ++q, ++q2) {
			P[n][q] = exp(H[n][q]-sup + (q2 == 0) * Hext[n]);
			norm += P[n][q];
		}

		for (q = 0, q2 = -Q; q2 <= Q; ++q, ++q2)
			P[n][q] /= norm;
	}

}

double entropy()
{
	int m, n, q, q2;
	double A, ca, mom1, mom2, C;
	double Ssigma , Smu , Snode =0, Ssite =0, ss, ee;
	double * u = auxv[omp_get_thread_num()];

	compute_marginals();
	for (n = 0; n < N; ++n) {
		for (q = 0, q2 = -Q; q2 <= Q; ++q, ++q2) { 
			if (P[n][q] > SMALL) 
				Ssite += P[n][q] * log(P[n][q]) * (M - 1);
		}
	}
	
	for (m = 0; m < M; ++m) {
		Smu = 0; 
		Ssigma= 0;
		for (n = 0; n < N; ++n) {
			mom1 = 0;
			mom2 = 0;
			u = get_message(m, n);
			for (q = 0, q2 = -Q; q2 <= Q; ++q, ++q2) {
				if (u[q] > SMALL) {
					Snode += P[n][q] * log( u[q] );
					mom1 += u[q] * q2;
					mom2 += u[q] * q2 * q2;
				}
			}
			Smu += mom1 * X[m][n];
			Ssigma += (mom2 - mom1 * mom1) * X[m][n] * X[m][n]; 
		}
		if (nonlinear) {
			A = normal(Smu, 1. / sqrt(Ssigma), 0);
			ca = A * Y[m] + (1 - A) * (1 - Y[m]);
			if (Y[m] * (1 - Y[m]))
				C = A * (Y[m] * log(Y[m])) + (1 - A) * ((1 - Y[m]) * log(1 - Y[m]));			
			else
				C =  0;
		} else {
			ss = (1 + 2 * beta * Ssigma * Ssigma);
			ee = exp(-beta * pow(Y[m]-Smu, 2) / ss);
			ca = ee / sqrt(ss);
			C = ee / pow(ss, 2.5) * (pow(Y[m] - Smu, 2) + Ssigma * Ssigma + 2 * pow(Ssigma, 4) * beta);
		}
				
		Snode += -log(ca) + C/ca;
	}
	return (-Snode + Ssite) / N;
}



int paracooling(int T, double deltahdil) 
{
	int t,n;
	double diff = 0, S;

	converge(T);
	for (t = 0; t < T; ++t) {
		diff = iterate();
		diff = iterate();
		diff = iterate();
		diff = iterate();
		diff = iterate();
		diff = iterate();

/* 		converge(T); */
		S = entropy();
		hdil += deltahdil;
		for(n=0; n<N; ++n ) 
			Hext[n] = hdil;

		if (S < 0)
			break;
		printf("paracool: %i %g %g %g\n", t, S, hdil, diff);
	}
	converge(T);
	return 0;
}

int comp(void const * a, void const * b) {
        struct val_index const *aa = a;
        struct val_index const *bb = b;
        if (aa->val > bb->val)
                return 1;
        if (aa->val < bb->val)
                return -1;
        return 0;
}


int decimate(int T, int deltan, int nmax)
{

	int n, start, stop, it=0;
	double S;
	struct val_index * vi;
	

	vi = (struct val_index * ) malloc ( N * sizeof( struct val_index ) );
	deltan = (int)N/2.;
	start = 0;
	stop = deltan;
	do{
		converge(T);
		compute_marginals();
		for ( n = 0; n < N; ++n ) {
			vi[n].index = n;
			vi[n].val = - P[n][Q]; /* probability to be zero */
		}
		qsort(vi, N , sizeof(struct val_index), &comp);
		for(n = start; n<stop; ++n ) 
			Hext[vi[n].index] = Hmax;
		start = stop;
		deltan = (int)(deltan/2.);
		stop += deltan;
		S = entropy();
		printf("decimation: %i %g\n", it, S);
		it++;
	}while ( stop <= nmax && S>0);
	free(vi);
	return(0);
}



